import { Readable, Updater, Unsubscriber } from 'svelte/store';
import { z } from 'zod';
import { ActionReturn } from 'svelte/action';

/**
 * Instance that hold all our field's state, as Svelte's Store
 *
 * @see {@link ZodFormStore}
 * @template K Name of the field
 * @template A Zod's parent schema type of the field
 * @template O Type of the parent data
 */
declare class ZodFieldStore<K extends Extract<keyof O, string>, A extends z.ZodRawShape, O = A> {
    /**
     * Field name
     */
    name: K;
    /**
     * Readable store holding field's value
     */
    value: Readable<O[K]>;
    /**
     * The field have been touched or not
     */
    touched: Readable<boolean>;
    /**
     * The field value been changed or not
     */
    dirty: Readable<boolean>;
    /**
     * The field validation error, if any
     */
    error: Readable<string>;
    /**
     * The field value is valid or not
     */
    valid: Readable<boolean>;
    /**
     * Callback to update field's value
     */
    updateValue: (updater: Updater<O[K]>) => void;
    /**
     * Function to set field's value
     */
    setValue: (val: O[K]) => void;
    /**
     * Callback to update field's value.
     * Should be used in input's `onInput` event
     *
     * If you want to pass value directly, please use `setValue` instead
     */
    handleChange: (e: Event) => void;
    /**
     * Callback to mark field as touched
     */
    handleBlur: () => void;
    /**
     * Reset field to original state
     */
    reset: () => void;
    /**
     * Set custom field error
     */
    setError: (e: string) => void;
    /**
     * Update touched state
     */
    setTouched: (v: boolean) => void;
    constructor(parentSchema: z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O>, name: K, initialValue?: unknown);
}

/**
 * Settings for ZodFormStore
 */
interface ZodFormStoreOptions<T> {
    /**
     * The initial value of fields in the form.
     */
    initialValue?: Partial<T>;
    /**
     * Async callback to handle submission of the form.
     * Should return nothing, or an `string` contain error message
     */
    onSubmit?: (v: T) => Promise<void | string> | string | void;
    /** Print debug messages */
    debug?: boolean;
}
/**
 * Instance that hold all our form's state, as Svelte's Store
 *
 * @template A Input Zod schema for data in the form
 * @template O Output type. Represents the type of data in the form
 */
declare class ZodFormStore<A extends z.ZodRawShape, O extends object> {
    /** Zod's schema for data in the form. Should be a `ZodObject`. */
    schema: z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O> | z.ZodEffects<z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O>>;
    /** Form's data. Will be passed to onSubmit handler */
    readonly model: Readable<O>;
    /** Form settings. Should not be update */
    readonly options: ZodFormStoreOptions<O>;
    /**
     * Generated fields's functions
     */
    fields: {
        [K in Extract<keyof O, string>]: ZodFieldStore<K, A, O>;
    };
    /** Generated fields's stores, for easier access */
    stores: {
        [K in Extract<keyof O, string> as `${K}_value`]: ZodFieldStore<K, A, O>['value'];
    } & {
        [K in Extract<keyof O, string> as `${K}_touched`]: ZodFieldStore<K, A, O>['touched'];
    } & {
        [K in Extract<keyof O, string> as `${K}_dirty`]: ZodFieldStore<K, A, O>['dirty'];
    } & {
        [K in Extract<keyof O, string> as `${K}_error`]: ZodFieldStore<K, A, O>['error'];
    } & {
        [K in Extract<keyof O, string> as `${K}_valid`]: ZodFieldStore<K, A, O>['valid'];
    };
    /**
     * Function to start parsing, validating and submit the form's data.
     *
     * Should assign to form's submit event or submit button.
     *
     * ```svelte
     * <form on:submit={form.triggerSubmit}>
     *  ...
     * </form>
     * ```
     */
    triggerSubmit: () => Promise<void>;
    /**
     * Function to reset the form to original state.
     *
     * Should assign to form's `reset` event, reset button, or call directly.
     *
     * ```svelte
     * <form on:reset={form.reset}>
     *  ...
     * </form>
     * ```
     */
    reset: () => void;
    /** Indicate if the form is being submitted (`onSubmit` handler is resolving).  */
    submitting: Readable<boolean>;
    /**
     * Array of string contains all error messages
     * (including fields's errors and error return from `onSubmit` handler).
     */
    errors: Readable<string[]>;
    /** Error message returned from `onSubmit` handler, or custom validation message. */
    error: Readable<string>;
    /** Indicate if the form is edited or submitted. */
    dirty: Readable<boolean>;
    /** Indicate if the form is valid. */
    valid: Readable<boolean>;
    constructor(
    /** Zod's schema for data in the form. Should be a `ZodObject`. */
    schema: z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O> | z.ZodEffects<z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O>>, 
    /** Additional configuration */
    options?: ZodFormStoreOptions<O>);
    /** Set the form's error message manually */
    setError(errorMessage: string, path: (string | number)[]): void;
    /**
     * Setup auto submit on every change of the model.
     * If the model changed, it will be submit in last `delay` milliseconds.
     *
     * @param delay Time in milliseconds to wait before submitting the form.
     */
    setupAutoSubmit(delay: number): Unsubscriber;
}

type TSupportedElement$3 = HTMLParagraphElement | HTMLSpanElement | HTMLDivElement;
/**
 * Connects an error feedback element to a ZodFieldStore, enabling automatic show/hide
 * of the error message with the field store.
 */
declare function connectError<K extends Extract<keyof O, string>, A extends z.ZodRawShape, O = A>(node: TSupportedElement$3, param: ZodFieldStore<K, A, O>): ActionReturn<ZodFieldStore<K, A, O>>;

type TSupportedElement$2 = HTMLInputElement | HTMLTextAreaElement;
/**
 * Connects an input element to a ZodFieldStore, enabling automatic synchronization
 * of the input's value with the store and handling of input events.
 *
 * Supports text inputs and textareas.
 */
declare function connectInput<K extends Extract<keyof O, string>, A extends z.ZodRawShape, O = A>(node: TSupportedElement$2, param: ZodFieldStore<K, A, O>): ActionReturn<ZodFieldStore<K, A, O>>;

type TSupportedElement$1 = HTMLSelectElement;
/**
 * Connects a select element to a ZodFieldStore, enabling automatic synchronization
 * of the select's value with the store and handling of input events.
 */
declare function connectSelect<K extends Extract<keyof O, string>, A extends z.ZodRawShape, O = A>(node: TSupportedElement$1, param: ZodFieldStore<K, A, O>): ActionReturn<ZodFieldStore<K, A, O>>;

type TSupportedElement = HTMLButtonElement;
/**
 * Connects a button element to a ZodFormStore, enabling automatic synchronization
 * of the button's disabled state with the store and handling of submit events.
 */
declare function connectSubmit<A extends z.ZodRawShape, O extends object = A>(node: TSupportedElement, param: ZodFormStore<A, O>): ActionReturn<ZodFormStore<A, O>>;

export { ZodFieldStore, ZodFormStore, type ZodFormStoreOptions, connectError, connectInput, connectSelect, connectSubmit };
