import { writable, readonly, derived, get } from 'svelte/store';
import { ZodError, ZodEffects } from 'zod';

const getErrorMessage = (e) => {
  if (e instanceof ZodError)
    return e.issues.filter((a) => a.path.length === 0).map((a) => a.message).join();
  if (e instanceof Error) return e.message;
  if (typeof e === "string") return e;
  if (typeof e === "undefined") return "";
  if (e === null) return "";
  return JSON.stringify(e);
};

class ZodFieldStore {
  /**
   * Field name
   */
  name;
  /**
   * Readable store holding field's value
   */
  value;
  /**
   * The field have been touched or not
   */
  touched;
  /**
   * The field value been changed or not
   */
  dirty;
  /**
   * The field validation error, if any
   */
  error;
  /**
   * The field value is valid or not
   */
  valid;
  /**
   * Callback to update field's value
   */
  updateValue;
  /**
   * Function to set field's value
   */
  setValue;
  /**
   * Callback to update field's value.
   * Should be used in input's `onInput` event
   *
   * If you want to pass value directly, please use `setValue` instead
   */
  handleChange;
  /**
   * Callback to mark field as touched
   */
  handleBlur;
  /**
   * Reset field to original state
   */
  reset;
  /**
   * Set custom field error
   */
  setError;
  /**
   * Update touched state
   */
  setTouched;
  constructor(parentSchema, name, initialValue) {
    if (!(name in parentSchema.shape)) throw new Error("Invalid field name");
    const touched = writable(false);
    const dirty = writable(false);
    const error = writable("");
    const value = writable();
    const schema = parentSchema.shape[name];
    const handleError = (e) => error.set(getErrorMessage(e));
    const resetValue = () => {
      value.set(initialValue);
      try {
        const parsed = schema.parse(initialValue);
        value.set(parsed);
      } catch (e) {
        error.set(getErrorMessage(e));
      }
    };
    resetValue();
    const handleChange = (e) => {
      error.set("");
      let nextVal = null;
      if (!(e instanceof Event)) throw new Error("Invalid param passes to `handleChange`. If you want to pass value directly, please use `setValue` instead" /* invalidEvent */, { cause: e });
      if (e.currentTarget instanceof HTMLInputElement)
        if (e.currentTarget.type === "checkbox") nextVal = e.currentTarget.checked;
        else nextVal = e.currentTarget.value;
      else if (e.currentTarget instanceof HTMLSelectElement || e.currentTarget instanceof HTMLTextAreaElement)
        nextVal = e.currentTarget.value;
      else throw new Error("Invalid event target. Supported elements are: HTMLInputElement, HTMLSelectElement, HTMLTextAreaElement" /* invalidEventTarget */, { cause: e });
      try {
        nextVal = schema.parse(nextVal);
      } catch (e2) {
        handleError(e2);
      }
      value.set(nextVal);
      touched.set(true);
      dirty.set(true);
    };
    const handleSetValue = (v) => {
      error.set("");
      if (schema)
        try {
          value.set(schema.parse(v));
        } catch (e) {
          handleError(e);
        }
      else value.set(v);
      touched.set(true);
      dirty.set(true);
    };
    const handleUpdateValue = (updater) => {
      const safeUpdater = (v) => {
        error.set("");
        const updatedVal = updater(v);
        if (schema)
          try {
            return schema.parse(updatedVal);
          } catch (e) {
            handleError(e);
          }
        return updatedVal;
      };
      value.update(safeUpdater);
      touched.set(true);
      dirty.set(true);
    };
    this.name = name;
    this.value = readonly(value);
    this.touched = readonly(touched);
    this.dirty = readonly(dirty);
    this.error = readonly(error);
    this.valid = derived(error, (e) => !e);
    this.handleChange = handleChange;
    this.updateValue = handleUpdateValue;
    this.setValue = handleSetValue;
    this.handleBlur = () => touched.set(true);
    this.setError = (e) => error.set(e);
    this.setTouched = (v) => touched.set(v);
    this.reset = () => {
      touched.set(false);
      dirty.set(false);
      error.set("");
      resetValue();
    };
  }
}

const debounce = ({ delay }, func) => {
  let timer = void 0;
  let active = true;
  const debounced = (...args) => {
    if (active) {
      clearTimeout(timer);
      timer = setTimeout(() => {
        if (active) func(...args);
        timer = void 0;
      }, delay);
    } else {
      func(...args);
    }
  };
  debounced.isPending = () => {
    return timer !== void 0;
  };
  debounced.cancel = () => {
    active = false;
  };
  debounced.flush = (...args) => func(...args);
  return debounced;
};

function zip(...arrays) {
  if (!arrays || !arrays.length) return [];
  return new Array(Math.max(...arrays.map(({ length }) => length))).fill([]).map((_, idx) => arrays.map((array) => array[idx]));
}

class ZodFormStore {
  constructor(schema, options) {
    this.schema = schema;
    this.options = options || {};
    const submitting = writable(false);
    const dirty = writable(false);
    const formError = writable("");
    const fieldNames = Object.keys(
      schema instanceof ZodEffects ? schema.innerType().shape : schema.shape
    );
    const fieldStores = fieldNames.map(
      (name) => new ZodFieldStore(
        schema instanceof ZodEffects ? schema.innerType() : schema,
        name,
        this.options.initialValue?.[name]
      )
    );
    const model = derived(
      fieldStores.map((z) => z.value),
      (z) => Object.fromEntries(zip(fieldNames, z))
    );
    const fieldErrors = derived(
      fieldStores.map((z) => z.error),
      (e) => e.filter(Boolean)
    );
    const triggerSubmit = async () => {
      dirty.set(true);
      fieldStores.forEach((f) => f.setTouched(true));
      if (!this.options.onSubmit) return;
      const _fieldErrors = get(fieldErrors);
      if (_fieldErrors.length > 0) {
        if (this.options.debug) console.debug("\u2757\uFE0F Form not valid", _fieldErrors);
        return;
      }
      formError.set("");
      submitting.set(true);
      try {
        const values = schema.parse(get(model));
        const result = await this.options.onSubmit(values);
        if (result) formError.set(result);
      } catch (e) {
        if (this.options.debug) console.debug(`\u{1F6AB} Form submit error`, e);
        if (e instanceof ZodError)
          for (const issue of e.issues) this.setError(issue.message, issue.path);
        formError.set(getErrorMessage(e));
      } finally {
        submitting.set(false);
      }
    };
    function handleReset() {
      dirty.set(false);
      submitting.set(false);
      formError.set("");
      fieldStores.forEach((f) => f.reset());
    }
    this.fields = Object.fromEntries(
      fieldStores.map((z) => [z.name, z])
    );
    const handlerFuncNames = [
      "updateValue",
      "setValue",
      "handleChange",
      "handleBlur",
      "reset",
      "setError",
      "setTouched"
    ];
    this.stores = Object.fromEntries(
      fieldStores.flatMap(
        (fieldStore) => Object.keys(fieldStore).filter((key) => key !== "name" && !handlerFuncNames.includes(key)).map((prop) => [
          `${String(fieldStore.name)}_${prop}`,
          fieldStore[prop]
        ])
      )
    );
    this.triggerSubmit = triggerSubmit;
    this.reset = handleReset;
    this.submitting = readonly(submitting);
    this.error = readonly(formError);
    this.dirty = derived([dirty, ...fieldStores.map((a) => a.dirty)], (a) => a.some((b) => b));
    this.errors = derived(
      [fieldErrors, formError],
      (errors) => errors.flatMap((z) => z).filter(Boolean)
    );
    this.model = model;
    this.valid = derived(this.errors, (e) => !e.length);
  }
  /** Form's data. Will be passed to onSubmit handler */
  model;
  /** Form settings. Should not be update */
  options;
  /**
   * Generated fields's functions
   */
  fields;
  /** Generated fields's stores, for easier access */
  // prettier-ignore
  stores;
  /**
   * Function to start parsing, validating and submit the form's data.
   *
   * Should assign to form's submit event or submit button.
   *
   * ```svelte
   * <form on:submit={form.triggerSubmit}>
   *  ...
   * </form>
   * ```
   */
  triggerSubmit;
  /**
   * Function to reset the form to original state.
   *
   * Should assign to form's `reset` event, reset button, or call directly.
   *
   * ```svelte
   * <form on:reset={form.reset}>
   *  ...
   * </form>
   * ```
   */
  reset;
  /** Indicate if the form is being submitted (`onSubmit` handler is resolving).  */
  submitting;
  /**
   * Array of string contains all error messages
   * (including fields's errors and error return from `onSubmit` handler).
   */
  errors;
  /** Error message returned from `onSubmit` handler, or custom validation message. */
  error;
  /** Indicate if the form is edited or submitted. */
  dirty;
  /** Indicate if the form is valid. */
  valid;
  /** Set the form's error message manually */
  setError(errorMessage, path) {
    const [currentPath] = path;
    const field = this.fields[currentPath];
    if (!field) return;
    field.setError(errorMessage);
  }
  /**
   * Setup auto submit on every change of the model.
   * If the model changed, it will be submit in last `delay` milliseconds.
   *
   * @param delay Time in milliseconds to wait before submitting the form.
   */
  setupAutoSubmit(delay) {
    const debouncedTriggerSubmit = debounce({ delay }, () => {
      if (this.options.debug) console.debug("\u23F1\uFE0F Auto submitting...");
      this.triggerSubmit();
    });
    let prevModel = null;
    const sub = this.model.subscribe((nextModel) => {
      if (prevModel === null) {
        prevModel = nextModel;
        return;
      }
      if (prevModel === nextModel) return;
      prevModel = nextModel;
      debouncedTriggerSubmit();
    });
    return sub;
  }
}

function connectError(node, param) {
  let cleanup = initialize$3(node, param);
  return {
    destroy: () => cleanup(),
    update: (nextParams) => {
      cleanup();
      cleanup = initialize$3(node, nextParams);
    }
  };
}
function initialize$3(node, param) {
  return derived([param.error, param.touched], ([a, b]) => [a, b]).subscribe(([error, touched]) => {
    if (typeof error === "string" && error && touched) {
      node.innerText = error;
      node.style.display = "";
    } else {
      node.innerText = "";
      node.style.display = "none";
    }
  });
}

function connectInput(node, param) {
  let cleanup = initialize$2(node, param);
  return {
    destroy: () => cleanup(),
    update: (nextParams) => {
      cleanup();
      cleanup = initialize$2(node, nextParams);
    }
  };
}
function initialize$2(node, store) {
  const { handleChange, handleBlur } = store;
  node.addEventListener("input", handleChange);
  node.addEventListener("blur", handleBlur);
  const unsubscribes = [
    // Sync value to node
    store.value.subscribe((v) => {
      if (typeof v === "undefined") node.value = "";
      else if (typeof v === "string") node.value = v;
      else if (v instanceof Date) node.value = v.toISOString();
      else console.log("\u26A0\uFE0F Unsupported type");
    }),
    // Set 'invalid' class
    derived([store.error, store.touched], ([a, b]) => [a, b]).subscribe(
      ([error, touched]) => node.classList.toggle("invalid", Boolean(error && touched))
    ),
    // Set 'valid' class
    derived([store.error, store.dirty], ([a, b]) => [a, b]).subscribe(
      ([error, dirty]) => node.classList.toggle("valid", Boolean(!error && dirty))
    )
  ];
  const cleanup = () => {
    unsubscribes.forEach((a) => a());
    node.removeEventListener("input", handleChange);
    node.removeEventListener("blur", handleBlur);
  };
  return cleanup;
}

function connectSelect(node, param) {
  let cleanup = initialize$1(node, param);
  return {
    destroy: () => cleanup(),
    update: (nextParams) => {
      cleanup();
      cleanup = initialize$1(node, nextParams);
    }
  };
}
function initialize$1(node, store) {
  const { value, handleChange, setValue, handleBlur } = store;
  const handleSelectChange = node.multiple ? (e) => {
    const elm = e.target;
    const val = Array.from(elm.selectedOptions).map((a) => a.value);
    setValue(val);
  } : handleChange;
  node.addEventListener("input", handleSelectChange);
  node.addEventListener("blur", handleBlur);
  const unsubscribes = [
    value.subscribe((v) => {
      if (typeof v === "undefined") node.value = "";
      else if (typeof v === "string") node.value = v;
      else if (Array.isArray(v)) for (const o of node.options) o.selected = v.includes(o.value);
    }),
    // Set 'invalid' class
    derived([store.error, store.touched], ([a, b]) => [a, b]).subscribe(
      ([error, touched]) => node.classList.toggle("invalid", Boolean(error && touched))
    ),
    // Set 'valid' class
    derived([store.error, store.dirty], ([a, b]) => [a, b]).subscribe(
      ([error, dirty]) => node.classList.toggle("valid", Boolean(!error && dirty))
    )
  ];
  const destroy = () => {
    unsubscribes.forEach((a) => a());
    node.removeEventListener("input", handleSelectChange);
    node.removeEventListener("blur", handleBlur);
  };
  return destroy;
}

function connectSubmit(node, param) {
  let cleanup = initialize(node, param);
  return {
    destroy: () => cleanup(),
    update: (nextParams) => {
      cleanup();
      cleanup = initialize(node, nextParams);
    }
  };
}
function initialize(node, store) {
  return derived([store.valid, store.submitting], ([a, b]) => [a, b]).subscribe(
    ([valid, submitting]) => {
      node.disabled = !valid || submitting;
    }
  );
}

export { ZodFormStore, connectError, connectInput, connectSelect, connectSubmit };
